"""Constants used across the feathersdk library"""


ROOT_FEATHER_PATH = "/feathersys"

USER_CONFIG_PATH = ROOT_FEATHER_PATH + "/user_config"
MOTORS_CALIBRATION_PATH = USER_CONFIG_PATH + "/motors_calibration"

STATE_PATH = ROOT_FEATHER_PATH + "/state"
POWER_STATE_PATH = STATE_PATH + "/power_state.json"
BATTERY_HISTORY_PATH = STATE_PATH + "/battery_history.json"

# Print line width for separator lines (e.g., "=" * PRINT_LINE_WIDTH)
PRINT_LINE_WIDTH = 70
